/* Copyright (C) 2001-2003 Peter Selinger.
   This file is part of potrace. It is free software and it is covered
   by the GNU general public license. See the file COPYING for details. */

/* $Id: pbm.c,v 1.6 2003/08/08 21:24:00 selinger Exp $ */

/* Routines for manipulating bitmaps, including reading pbm files. */

#include <stdlib.h>
#include <errno.h>
#include <string.h>

#include "pbm.h"

/* ---------------------------------------------------------------------- */
/* basic bitmap routines */

/* return new un-initialized bitmap. NULL with errno on error */

bitmap_t *bm_new(int w, int h) {
  bitmap_t *bm;
  int c = 1+(w-1)/WORDBITS;
  int errno_save;

  bm = malloc(sizeof(bitmap_t));
  if (!bm) {
    return NULL;
  }
  bm->w = w;
  bm->h = h;
  bm->c = c;
  bm->map = malloc(c*h*WORDSIZE);
  if (!bm->map) {
    errno_save = errno;
    free(bm);
    errno = errno_save;
    return NULL;
  }
  return bm;
}

/* free the given bitmap */
void bm_free(bitmap_t *bm) {
  if (bm) {
    free(bm->map);
  }
  free(bm);
}

/* duplicate the given bitmap. Return NULL on error with errno set. */
bitmap_t *bm_dup(bitmap_t *bm) {
  bitmap_t *bm1 = bm_new(bm->w, bm->h);
  if (!bm1) {
    return NULL;
  }
  memcpy(bm1->map, bm->map, bm->c*bm->h*WORDSIZE);
  return bm1;
}

/* clear the given bitmap. */
void bm_clear(bitmap_t *bm) {
  memset(bm->map, 0, bm->c*bm->h*WORDSIZE);
}

/* ---------------------------------------------------------------------- */
/* routines for reading pbm streams */

/* skip whitespace and comments, then read a non-negative decimal
   number from a stream. Return -1 on EOF. Tolerate other errors (skip
   bad characters). Also read the immediately following character, or
   to the end of the line if next character is a comment character. */

static int readnum(FILE *f) {
  int c;
  int acc;

  /* skip whitespace and comments */
  while (1) {
    c = fgetc(f);
    if (c=='#') {
      while (1) {
	c = fgetc(f);
	if (c=='\n' || c==EOF) {
	  break;
	}
      }
    }
    if (c==EOF) {
      return -1;
    }
    if (c>='0' && c<='9') {
      break;
    }
  }

  /* first digit is already in c */
  acc = c-'0';
  while (1) {
    c = fgetc(f);
    if (c=='#') {
      while (1) {
	c = fgetc(f);
	if (c=='\n' || c==EOF) {
	  break;
	}
      }
    }
    if (c<'0' || c>'9') {
      break;
    }
    acc *= 10;
    acc += c-'0';
  }
  return acc;
}

/* similar to readnum, but read only a single 0 or 1, and do not read
   any characters after it. */

static int readbit(FILE *f) {
  int c;

  /* skip whitespace and comments */
  while (1) {
    c = fgetc(f);
    if (c=='#') {
      while (1) {
	c = fgetc(f);
	if (c=='\n' || c==EOF) {
	  break;
	}
      }
    }
    if (c==EOF) {
      return -1;
    }
    if (c>='0' && c<='1') {
      break;
    }
  }

  /* first digit is already in c */
  return c-'0';
}

/* read a pbm stream: either P1 or P4 format. See pbm(5). Return
   bitmap in *bmp. Return 0 on success, -1 on error with errno set, -2
   on bad file format, and 1 on premature end of file, -3 if eof occurs
   immediately. If the return value is >=0, *bmp is valid. */

int bm_readpbm(FILE *f, bitmap_t **bmp) {
  int magic[2];
  bitmap_t *bm;
  int x, y, i, b;
  int bpr; /* bytes per row (as opposed to 4*bm->c) */
  int w, h;

  /* read header */
  magic[0] = fgetc(f);
  if (magic[0] == EOF) {
    return -3;
  } else if (magic[0] != 'P') {
    return -2;
  }

  magic[1] = fgetc(f);
  if (magic[1] != '1' && magic[1] != '4') {
    return -2;
  }

  w = readnum(f);
  if (w<0) {
    return -2;
  }

  h = readnum(f);
  if (h<0) {
    return -2;
  }

  /* allocate bitmap */
  bm = bm_new(w, h);
  if (!bm) {
    return -1;
  }

  /* zero it out */
  bm_clear(bm);

  switch (magic[1]) {
  case '1':
  default:
    /* read P1 format */
    
    for (y=h-1; y>=0; y--) {
      for (x=0; x<w; x++) {
	b = readbit(f);
	if (b<0) {
	  goto eof;
	}
	BM_UPUT(bm, x, y, b);
      }
    }
    break;

  case '4':
    /* read P4 format */
    
    bpr = 1+(w-1)/8;

    for (y=h-1; y>=0; y--) {
      for (i=0; i<bpr; i++) {
	b = fgetc(f);
	if (b==EOF) {
	  goto eof;
	}
	*bm_index(bm, i*8, y) |= (b << (8*(WORDSIZE-1-(i % WORDSIZE))));
      }
    }
    break;
  }

  *bmp = bm;
  return 0;

 eof:
  *bmp = bm;
  return 1;
}

/* ---------------------------------------------------------------------- */
/* output - for primitive debugging purposes only! */

/* print bitmap to screen */
int bm_print(FILE *f, bitmap_t *bm) {
  int x, y;
  int xx, yy;
  int d;
  int sw, sh;

  sw = bm->w < 79 ? bm->w : 79;
  sh = bm->w < 79 ? bm->h : bm->h*sw*44/(79*bm->w);

  for (yy=sh-1; yy>=0; yy--) {
    for (xx=0; xx<sw; xx++) {
      d=0;
      for (x=xx*bm->w/sw; x<(xx+1)*bm->w/sw; x++) {
	for (y=yy*bm->h/sh; y<(yy+1)*bm->h/sh; y++) {
	  if (BM_GET(bm, x, y)) {
	    d++;
	  }
	}
      }
      fputc(d ? '*' : ' ', f);
    }
    fputc('\n', f);
  }
  return 0;
}
